if (!/\[native code\]/.test(Object.create)) {
	Object.create = function (a) {
		var f = function () {
		}
		f.prototype = a;
		return new f()
	}
}


// import "babel-polyfill";

import Yox from 'yox';
import jQuery from 'jquery';
import _ from 'lodash';
import {Store} from "./lib/store"
import {Router} from "./lib/router"
import {dayjs} from "./lib/dayjs"

import pkg from '../package'

let $ = window.jQuery = window.$ = jQuery;


Yox.prototype.getSlot = function (name) {
	return this.get(`$slot_${name}`);
};

Yox.prototype.hasSlot = function (name) {
	return typeof this.getSlot(name) !== "undefined";
};


let FireBird = (function () {
	function FireBird(options) {
		var app = new Yox(options);
		return app;
	}

	return FireBird;
})();

_.each(Yox, (func, name) => {
	FireBird[name] = func;
});

FireBird.log = (msg) => {
	console.log(`[FireBird info]: ${msg}`);
};
FireBird.warn = (msg) => {
	console.warn(`[FireBird warn]: ${msg}`);
};
FireBird.error = (msg) => {
	console.error(`[FireBird error]: ${msg}`);
};


FireBird.components = {};

FireBird.component = (name, options) => {
	if (typeof name === "string") {
		if (typeof options === "function") {

		} else {
			options.name = name;
			if (options.extend) {

				if (typeof options.extend === "string") {
					options.extend = FireBird.components[options.extend] || null;
				}

				options = $.extend(true, {}, options.extend, options, {
					"_parent": options.extend.name
				});
			}
		}

		Yox.component(name, options);
		FireBird.log(`组件 ${name} 已注册`);
		FireBird.components[name] = options;
		return options;
	}
	if (typeof name === "object") {
		_.each(name, (options, comp) => {
			FireBird.component(comp, options);
		});
	}
};

FireBird.partials = {};

FireBird.partial = (name, template) => {
	if (typeof name === "string") {
		Yox.partial(name, template);
		FireBird.log(`模版 ${name} 已注册`);
		FireBird.partials[name] = template;
		return template;
	}
	if (typeof name === "object") {
		_.each(name, (template, partial) => {
			FireBird.partial(partial, template);
		});
	}
};

/**
 * 组件 class 处理器
 * @param cssObject
 * @returns {string}
 */
FireBird.css = (cssObject = []) => {
	let classes = [];
	_.each(cssObject, (css) => {
		if (_.isObject(css)) {
			_.each(css, (v, k) => {
				if (typeof v !== "undefined" && v === true) {
					classes.push(k);
				}
			});
		}
		if (_.isString(css)) {
			classes.push(css);
		}
	});
	return classes.join(" ");
};

FireBird.style = (styleObject = {}) => {
	let style = [];

	_.each(styleObject, (v, k) => {
		style.push(`${k}: ${v}`);
	});

	return style.join("; ");
};

FireBird.oneOf = (value, defaults, array = []) => {
	if (!FireBird.array.has(array, value)) {
		return defaults;
	}
	return value;
};


// 将 lodash 注入
FireBird.filter(_);

Store.install(FireBird);
Router.install(FireBird);

FireBird.component("PageApp", {
	template: "<div></div>",
	data() {
		return {};
	}
});


window.FireBird = FireBird;

let version = pkg.version;
console.log(" _______________________________________________________________________");
console.log("|                                                                       |");
console.log("|  ######## #### ########  ######## ########  #### ########  ########   |");
console.log("|  ##        ##  ##     ## ##       ##     ##  ##  ##     ## ##     ##  |");
console.log("|  ##        ##  ##     ## ##       ##     ##  ##  ##     ## ##     ##  |");
console.log("|  ######    ##  ########  ######   ########   ##  ########  ##     ##  |");
console.log("|  ##        ##  ##   ##   ##       ##     ##  ##  ##   ##   ##     ##  |");
console.log("|  ##        ##  ##    ##  ##       ##     ##  ##  ##    ##  ##     ##  |");
console.log("|  ##       #### ##     ## ######## ########  #### ##     ## ########   |");
console.log("|                                                                       |");
console.log(`|${_.pad("", 71)}|`);
console.log(`|${_.pad("Version: " + version, 71)}|`);
console.log(`|${_.pad("Date: " + dayjs().format(), 71)}|`);
console.log(`|${_.pad("Author: lincong1987@gmail.com", 71)}|`);
//console.log("|                                                                       |");
console.log("|_______________________________________________________________________|");


export {
	FireBird, jQuery, $, Router, Store, dayjs, version, _
};
                                                  
// exports.FireBird = FireBird;
// exports.Store = Store;
// exports.Router = Router;
// exports.jQuery = jQuery;
// exports.$ = $;
// exports._ = _;
// exports.dayjs = dayjs;
// exports.version = version;
